import inspect
from asyncio import iscoroutinefunction
from collections.abc import Awaitable, Callable, Coroutine
from dataclasses import dataclass
from typing import Any, Protocol, TypeAlias, TypeGuard

from fastapi import Depends
from htmy import Component, as_component_type

from holm.fastapi import FastAPIDependency


class SyncLayout(Protocol):
    """
    Sync layout protocol definition.
    """

    def __call__(self, __props: Any, /, **dependencies: Any) -> Any: ...


class AsyncLayout(Protocol):
    """
    Async layout protocol definition.
    """

    async def __call__(self, __props: Any, /, **dependencies: Any) -> Any: ...


Layout: TypeAlias = SyncLayout | AsyncLayout
"""
Layout type definition.

A layout is a callable that expects a single positional argument the component or data returned
by the wrapped layout or page, and returns the properties for its wrapper layout. The root layout
must always return a `Component`.
"""

LayoutFactory: TypeAlias = Callable[[Any], Any | Coroutine[None, None, Any]]
"""
A layout factory is a sync or asynccallable that expects a single argument (the layout's properties)
and returns the properties for its wrapper layout.

The root layout factory must always return a `htmy` `Component`.
"""


class LayoutDefinition(Protocol):
    """Protocol definition for objects (usually modules) that define a layout."""

    @property
    def layout(self) -> Layout: ...


def is_layout_definition(obj: Any) -> TypeGuard[LayoutDefinition]:
    """Type guard for `LayoutDefinition`."""
    layout = getattr(obj, "layout", None)
    return callable(layout)


@dataclass(frozen=True, slots=True)
class without_layout:
    """
    Marker that signals that the component it contains should not be
    wrapped by the application in any parent layouts.

    This utility must only be used as a return value wrapper in application
    components whose return value could be wrapped in a parent layout by
    `holm`. For example pages, submit handlers, layouts.

    Attempting to render an instance of this class results in a `RuntimeError`.
    The only reason it implements the `htmy()` method is to make static code
    analysis tools accept it as an `htmy.Component`.
    """

    component: Component
    """The wrapped component."""

    def htmy(self, _: Any) -> Component:
        raise RuntimeError(f"{type(self).__name__} must never be part of the htmy component tree!")


def combine_layouts_to_dependency(
    outer_dep: FastAPIDependency[LayoutFactory],
    inner: Layout | None,
) -> FastAPIDependency[LayoutFactory]:
    """
    Creates a layout factory dependency that wraps the inner layout inside the outer layout.

    Arguments:
        outer_dep: FastAPI dependency that returns the layout factory of the outer layout.
        inner: The inner layout component.
    """
    if inner is None:
        return outer_dep

    inner_dep = layout_to_dependency(inner)

    async def combined_layout_dep(
        outer: LayoutFactory = Depends(outer_dep),  # noqa: B008
        inner: LayoutFactory = Depends(inner_dep),  # noqa: B008
    ) -> LayoutFactory:
        async def layout_factory(props: Any) -> Any:
            inner_result = inner(props)
            if isinstance(inner_result, Awaitable):
                inner_result = await inner_result

            if isinstance(inner_result, without_layout):
                return inner_result.component

            result = outer(as_component_type(inner_result))
            if isinstance(result, Awaitable):
                result = await result

            return result

        return layout_factory

    return combined_layout_dep


def empty_layout(props: Any) -> Any:
    """`LayoutFactory` that simply returns the received props."""
    return props


def empty_layout_dependency() -> FastAPIDependency[LayoutFactory]:
    """
    FastAPI dependency that returns `empty_layout`.
    """
    return empty_layout


def layout_to_dependency(layout: Layout) -> FastAPIDependency[LayoutFactory]:
    """
    Converts a layout to a FastAPI dependency that returns a layout factory.
    """
    # eval_str=True is necessary in case future __annotations__ is used
    # where the layout is function is defined.
    params = tuple(inspect.signature(layout, eval_str=True).parameters.values())
    props_index = next((i for i, p in enumerate(params) if p.name != "self"), -1)
    if props_index == -1:
        raise ValueError("Layout factory must at least one argument.")

    result: FastAPIDependency[LayoutFactory]
    if iscoroutinefunction(layout):

        async def async_layout_factory_dep(**kwargs: Any) -> LayoutFactory:
            async def layout_factory(props: Any) -> Any:
                return await layout(props, **kwargs)

            return layout_factory

        result = async_layout_factory_dep
    else:

        def sync_layout_factory_dep(**kwargs: Any) -> LayoutFactory:
            def layout_factory(props: Any) -> Any:
                return layout(props, **kwargs)

            return layout_factory

        result = sync_layout_factory_dep

    # Copy the dependencies part of the signature from the original function, so
    # FastAPI can resolve those arguments. The props argument must be excluded.
    result.__signature__ = inspect.Signature(params[props_index + 1 :])  # type: ignore[union-attr]

    return result



import sys
from collections.abc import Callable, Collection, Sequence
from dataclasses import dataclass
from enum import Enum
from typing import Any, Literal, TypeAlias, TypeGuard

from fastapi.params import Depends as DependsParam

from holm.fastapi import FastAPIDependency

from ._metadata import MetadataMappingOrDependency

ActionDependency: TypeAlias = FastAPIDependency[Any]
"""
FastAPI action dependency.

Actions must normally return a `Component` or a FastAPI `Response`, but if they are wrapped
in a layout, then they must be allowed to return whatever their parent layout expects.
"""

ActionDependencyDecorator: TypeAlias = Callable[[ActionDependency], ActionDependency]


@dataclass(slots=True, eq=False, kw_only=True)
class ActionDescriptor:
    action: ActionDependency
    """FastAPI action dependency, basically the path operation."""

    route_args: dict[str, Any]
    """Arguments to pass to the created FastAPI route."""

    use_layout: bool
    """Whether the action should be rendered with all wrapping layouts."""

    metadata: MetadataMappingOrDependency | None
    """
    Optional metadata declaration for the action. The equivalent of page metadata for actions.

    The name of the attribute means `get_metadata_dependency()` can be used
    to resolve the attribute's value to a metadata dependency.
    """


ActionDescriptors: TypeAlias = dict[str, ActionDescriptor]
"""
Typing for the value of `_actions_variable`.

Dictionary that maps action paths to action descriptors. This allows
decorator stacking (registering the same action with multiple paths).
"""

HTTPMethod: TypeAlias = Literal["GET", "POST", "PUT", "PATCH", "DELETE"]


_actions_variable = "_holm_action_descriptors"
"""The name of the variable that stores action descriptors in modules."""


def get_actions(obj: Any) -> ActionDescriptors | None:
    """
    Loads and returns the action descriptors mapping if it exists in the given object.

    Action owners must have a variable whose name must match the value of `_actions_variable`,
    and the value must be a dictionary mapping action paths to action descriptors.
    """
    actions = getattr(obj, _actions_variable, None)
    return actions if isinstance(actions, dict) else None


def has_actions(obj: Any) -> TypeGuard[Any]:
    """
    "Type guard" that returns `True` if the given object contains actions.
    """
    return bool(get_actions(obj))


class action:
    """
    Decorators for registering actions.

    Decorator arguments map directly to the corresponding FastAPI route decorator's arguments,
    unless the documentation states otherwise.
    """

    @classmethod
    def __call__(
        cls,
        path: str | None = None,
        *,
        use_layout: bool = False,
        metadata: MetadataMappingOrDependency | None = None,
        methods: Collection[HTTPMethod] | None = None,
        dependencies: Sequence[DependsParam] | None = None,
        deprecated: bool | None = None,
        tags: list[str | Enum] | None = None,
    ) -> ActionDependencyDecorator:
        """
        Decorator for registering an action with arbitrary HTTP `methods`.

        Corresponding FastAPI route decorator: `APIRouter.api_route()`.

        Arguments:
            use_layout: Whether to render the action's return value in the
                containing package's layout (same as a page).
            metadata: Metadata mapping or dependency for the action. The equivalent
                of page metadata for actions.
            path: The action's path (within the package's `APIRouter`). If `None`,
                the decorated function's name is used as the path.
        """

        def decorator(func: ActionDependency) -> ActionDependency:
            cls._register_action(
                func,
                use_layout=use_layout,
                metadata=metadata,
                path=path,
                dependencies=dependencies,
                deprecated=deprecated,
                methods=methods,
                tags=tags,
            )
            return func

        return decorator

    @classmethod
    def get(
        cls,
        path: str | None = None,
        *,
        use_layout: bool = False,
        metadata: MetadataMappingOrDependency | None = None,
        dependencies: Sequence[DependsParam] | None = None,
        deprecated: bool | None = None,
        tags: list[str | Enum] | None = None,
    ) -> ActionDependencyDecorator:
        """
        Decorator for registering an action with an HTTP GET method.

        Corresponding FastAPI route decorator: `APIRouter.get()`.

        Arguments:
            use_layout: Whether to render the action's return value in the
                containing package's layout (same as a page).
            metadata: Metadata mapping or dependency for the action. The equivalent
                of page metadata for actions.
            path: The action's path (within the package's `APIRouter`). If `None`,
                the decorated function's name is used as the path.
        """

        def decorator(func: ActionDependency) -> ActionDependency:
            cls._register_action(
                func,
                use_layout=use_layout,
                metadata=metadata,
                path=path,
                dependencies=dependencies,
                deprecated=deprecated,
                methods=("GET",),
                tags=tags,
            )
            return func

        return decorator

    @classmethod
    def post(
        cls,
        path: str | None = None,
        *,
        use_layout: bool = False,
        metadata: MetadataMappingOrDependency | None = None,
        dependencies: Sequence[DependsParam] | None = None,
        deprecated: bool | None = None,
        tags: list[str | Enum] | None = None,
    ) -> ActionDependencyDecorator:
        """
        Decorator for registering an action with an HTTP POST method.

        Corresponding FastAPI route decorator: `APIRouter.post()`.

        Arguments:
            use_layout: Whether to render the action's return value in the
                containing package's layout (same as a page).
            metadata: Metadata mapping or dependency for the action. The equivalent
                of page metadata for actions.
            path: The action's path (within the package's `APIRouter`). If `None`,
                the decorated function's name is used as the path.
        """

        def decorator(func: ActionDependency) -> ActionDependency:
            cls._register_action(
                func,
                use_layout=use_layout,
                metadata=metadata,
                path=path,
                dependencies=dependencies,
                deprecated=deprecated,
                methods=("POST",),
                tags=tags,
            )
            return func

        return decorator

    @classmethod
    def put(
        cls,
        path: str | None = None,
        *,
        use_layout: bool = False,
        metadata: MetadataMappingOrDependency | None = None,
        dependencies: Sequence[DependsParam] | None = None,
        deprecated: bool | None = None,
        tags: list[str | Enum] | None = None,
    ) -> ActionDependencyDecorator:
        """
        Decorator for registering an action with an HTTP PUT method.

        Corresponding FastAPI route decorator: `APIRouter.put()`.

        Arguments:
            use_layout: Whether to render the action's return value in the
                containing package's layout (same as a page).
            metadata: Metadata mapping or dependency for the action. The equivalent
                of page metadata for actions.
            path: The action's path (within the package's `APIRouter`). If `None`,
                the decorated function's name is used as the path.
        """

        def decorator(func: ActionDependency) -> ActionDependency:
            cls._register_action(
                func,
                use_layout=use_layout,
                metadata=metadata,
                path=path,
                dependencies=dependencies,
                deprecated=deprecated,
                methods=("PUT",),
                tags=tags,
            )
            return func

        return decorator

    @classmethod
    def patch(
        cls,
        path: str | None = None,
        *,
        use_layout: bool = False,
        metadata: MetadataMappingOrDependency | None = None,
        dependencies: Sequence[DependsParam] | None = None,
        deprecated: bool | None = None,
        tags: list[str | Enum] | None = None,
    ) -> ActionDependencyDecorator:
        """
        Decorator for registering an action with an HTTP PATCH method.

        Corresponding FastAPI route decorator: `APIRouter.patch()`.

        Arguments:
            use_layout: Whether to render the action's return value in the
                containing package's layout (same as a page).
            metadata: Metadata mapping or dependency for the action. The equivalent
                of page metadata for actions.
            path: The action's path (within the package's `APIRouter`). If `None`,
                the decorated function's name is used as the path.
        """

        def decorator(func: ActionDependency) -> ActionDependency:
            cls._register_action(
                func,
                use_layout=use_layout,
                metadata=metadata,
                path=path,
                dependencies=dependencies,
                deprecated=deprecated,
                methods=("PATCH",),
                tags=tags,
            )
            return func

        return decorator

    @classmethod
    def delete(
        cls,
        path: str | None = None,
        *,
        use_layout: bool = False,
        metadata: MetadataMappingOrDependency | None = None,
        dependencies: Sequence[DependsParam] | None = None,
        deprecated: bool | None = None,
        tags: list[str | Enum] | None = None,
    ) -> ActionDependencyDecorator:
        """
        Decorator for registering an action with an HTTP DELETE method.

        Corresponding FastAPI route decorator: `APIRouter.delete()`.

        Arguments:
            use_layout: Whether to render the action's return value in the
                containing package's layout (same as a page).
            metadata: Metadata mapping or dependency for the action. The equivalent
                of page metadata for actions.
            path: The action's path (within the package's `APIRouter`). If `None`,
                the decorated function's name is used as the path.
        """

        def decorator(func: ActionDependency) -> ActionDependency:
            cls._register_action(
                func,
                use_layout=use_layout,
                metadata=metadata,
                path=path,
                dependencies=dependencies,
                deprecated=deprecated,
                methods=("DELETE",),
                tags=tags,
            )
            return func

        return decorator

    @classmethod
    def _register_action(
        cls,
        action: ActionDependency,
        *,
        use_layout: bool,
        metadata: MetadataMappingOrDependency | None,
        path: str | None,
        tags: list[str | Enum] | None,
        **route_args: Any,
    ) -> None:
        """
        Registers the given action in the module that contains it.

        Arguments:
            use_layout: Whether to render the action's return value in the
                containing package's layout (same as a page).
            metadata: Metadata mapping or dependency for the action. The equivalent
                of page metadata for actions.
            path: The action's path (within the package's `APIRouter`). If `None`,
                the decorated function's name is used as the path.
        """
        action_module = action.__module__
        action_name = action.__name__
        module = sys.modules[action_module]
        module_actions: ActionDescriptors | None = getattr(module, _actions_variable, None)
        if module_actions is None:
            module_actions = {}
            setattr(module, _actions_variable, module_actions)

        if path is None:
            path = f"/{action_name}"

        route_args["name"] = f"{action_module}.{action_name}"
        # Use the Action tag if no tags were set by the user.
        route_args["tags"] = ["Action"] if tags is None else tags
        route_args["response_model"] = None  # Don't generate a response schema.

        # Register the action.
        module_actions[path] = ActionDescriptor(
            action=action,
            route_args=route_args,
            use_layout=use_layout,
            metadata=metadata,
        )
